from typing import Optional, Union

from ._compat import (
    SCRIPT_RUN_CONTEXT_ATTR_NAME,
    AppSession,
    AppSessionState,
    ScriptRunContext,
)


def get_app_context(session: AppSession) -> Union[ScriptRunContext, None]:
    # HACK: Get ScriptRunContext from AppSession via ScriptThread
    scriptrunner = session._scriptrunner
    if not scriptrunner:
        return None
    if not scriptrunner._script_thread:
        return None

    script_thread = scriptrunner._script_thread
    ctx: Optional[ScriptRunContext] = getattr(
        script_thread, SCRIPT_RUN_CONTEXT_ATTR_NAME
    )
    return ctx


def is_rerunnable(session: AppSession) -> bool:
    ctx = get_app_context(session)
    if not hasattr(ctx, "_has_script_started"):
        # `ctx._has_script_started` has been introduced since 0.84.2
        # Ref: https://github.com/streamlit/streamlit/compare/0.84.1...0.84.2
        # Ref: https://github.com/streamlit/streamlit/pull/3550
        return True

    if ctx and not ctx._has_script_started:
        # This case is mainly when called from inside callbacks.
        # Callbacks are called after ctx is set and
        # before ctx._has_script_started is set as True.
        # Rel: https://github.com/whitphx/streamlit-server-state/issues/37
        return False
    if session._state == AppSessionState.SHUTDOWN_REQUESTED:
        # This case has no meaning on rerunning and causes an error
        # "Discarding ScriptRequest.RERUN request after shutdown".
        return False
    return True



import collections.abc
import threading
from typing import Iterator

from streamlit_server_state.server_state import ServerState


class ServerStateLock(collections.abc.Mapping):
    _server_state: ServerState

    def __init__(self, server_state: ServerState) -> None:
        self._server_state = server_state

    def __getitem__(self, k: str) -> threading.RLock:
        item = self._server_state._ensure_item(k)

        return item._value_lock

    def __getattr__(self, k: str) -> threading.RLock:
        return self.__getitem__(k)

    def __contains__(self, k: object) -> bool:
        return k in self._server_state.__items__

    def __iter__(self) -> Iterator[str]:
        return (k for k in self._server_state.__items__.keys())

    def __len__(self) -> int:
        return self._server_state.__items__.__len__()



import logging

from .app_context import get_app_context
from .session_info import get_this_session

logger = logging.getLogger(__name__)

_SERVER_STATE_RERUN_SUPPRESSION_ATTR_NAME_ = "__SERVER_STATE_SUPPRESS_RERUN__"


class RerunSuppressor:
    def __enter__(self) -> None:
        logger.debug("Start suppressing rerunning")

        # Context managers used in Streamlit apps should be managed
        # being bound to the app context object
        # following the design of the built-in components
        # that are used in `with` statements.
        # Ref: https://github.com/streamlit/streamlit/blob/1.15.0/lib/streamlit/delta_generator.py#L282-L300 # noqa: E501
        this_session = get_this_session()
        ctx = get_app_context(this_session)

        setattr(ctx, _SERVER_STATE_RERUN_SUPPRESSION_ATTR_NAME_, True)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        this_session = get_this_session()
        ctx = get_app_context(this_session)

        delattr(ctx, _SERVER_STATE_RERUN_SUPPRESSION_ATTR_NAME_)

        logger.debug("Finished suppressing rerunning")


no_rerun = RerunSuppressor()


def is_rerun_suppressed() -> bool:
    try:
        this_session = get_this_session()
    except RuntimeError:
        return False

    ctx = get_app_context(this_session)

    rerun_suppressed = getattr(ctx, _SERVER_STATE_RERUN_SUPPRESSION_ATTR_NAME_, False)
    logger.debug("Check the rerun-suppression flag: %r", rerun_suppressed)

    return rerun_suppressed



import collections.abc
from typing import Any, Dict, Iterator

from .server_state_item import ServerStateItem, ValueNotSetError
from .session_info import get_this_session


class ServerState(collections.abc.MutableMapping):
    # NOTE: This is initialized only once across all the instances of this class.
    # It is not a problem because the instance is singleton.
    # However, it can be in tests where multiple instances are created.
    # So, `self.__items__.clear()` is called inside `__init__` just for tests.
    __items__: Dict[str, ServerStateItem] = {}

    def __init__(self) -> None:
        super().__init__()
        self.__items__.clear()

    def _ensure_item(self, k: str) -> ServerStateItem:
        if k in self.__items__:
            item = self.__items__[k]
        else:
            item = ServerStateItem()
            self.__items__[k] = item

        return item

    def _ensure_item_in_this_session(self, k: str) -> ServerStateItem:
        item = self._ensure_item(k)

        this_session = get_this_session()

        item.bind_session(this_session)

        return item

    def __setitem__(self, k: str, v: Any) -> None:
        if k == "__items__":
            raise KeyError(f'Attr name "{k}" is forbidden')

        item = self._ensure_item(k)
        item.set_value(v)

    def __getitem__(self, k: str) -> Any:
        if k == "__items__":
            raise KeyError(f'Attr name "{k}" is forbidden')

        item = self._ensure_item_in_this_session(k)

        try:
            return item.get_value()
        except ValueNotSetError:
            raise KeyError(k)

    def __delitem__(self, k: str) -> None:
        if k == "__items__":
            raise KeyError(f'Attr name "{k}" is forbidden')

        del self.__items__[k]

    def __setattr__(self, name: str, value: Any) -> None:
        return self.__setitem__(name, value)

    def __getattr__(self, k: str) -> Any:
        return self.__getitem__(k)

    def __delattr__(self, k) -> None:
        return self.__delitem__(k)

    def __contains__(self, k: object) -> bool:
        if not isinstance(k, str):
            return False

        item = self._ensure_item_in_this_session(k)

        return item._is_set

    def __iter__(self) -> Iterator[str]:
        return (k for k, _ in ((k, v) for k, v in self.__items__.items() if v._is_set))

    def __len__(self) -> int:
        return len([i for i in self.__items__.values() if i._is_set])



import threading
import weakref
from typing import Generic, Optional, TypeVar

from .app_context import AppSession, is_rerunnable
from .hash import Hash, calc_hash
from .rerun_suppression import is_rerun_suppressed

StateValueT = TypeVar("StateValueT")


class ValueNotSetError(Exception):
    pass


class ServerStateItem(Generic[StateValueT]):
    _is_set: bool
    _value: StateValueT
    _value_hash: Optional[Hash]
    _value_lock: threading.RLock

    _bound_sessions: "weakref.WeakSet[AppSession]"
    _bound_sessions_lock: threading.Lock

    def __init__(self) -> None:
        self._is_set = False
        self._value_hash = None
        self._value_lock = threading.RLock()

        self._bound_sessions = weakref.WeakSet()
        self._bound_sessions_lock = threading.Lock()

    def bind_session(self, session: AppSession) -> None:
        with self._bound_sessions_lock:
            if session not in self._bound_sessions:
                self._bound_sessions.add(session)

    def _rerun_bound_sessions(self) -> None:
        with self._bound_sessions_lock:
            for session in self._bound_sessions:
                self._rerun_session_if_possible(session)

    def _rerun_session_if_possible(self, session: AppSession) -> None:
        if is_rerunnable(session):
            session.request_rerun(client_state=None)  # HACK: XD

    def _on_set(self):
        new_value_hash = calc_hash(self._value)
        if not is_rerun_suppressed():
            if self._value_hash is None or self._value_hash != new_value_hash:
                self._rerun_bound_sessions()

        self._value_hash = new_value_hash

    def set_value(self, value: StateValueT) -> None:
        with self._value_lock:
            self._is_set = True
            self._value = value

        self._on_set()

    def get_value(self) -> StateValueT:
        with self._value_lock:
            if not self._is_set:
                raise ValueNotSetError
            return self._value



import logging

from .rerun import make_force_rerun_bound_sessions
from .rerun_suppression import no_rerun
from .server import get_current_server
from .server_state import ServerState as _ServerState
from .server_state_lock import ServerStateLock as _ServerStateLock

logger = logging.getLogger(__name__)

_SERVER_STATE_KEY_ = "_server_state"
_SERVER_STATE_LOCK_KEY_ = "_server_state_lock"

_server = None

try:
    _server = get_current_server()
except RuntimeError as e:
    # NOTE: An error can be raised from the line above
    # when the app script uses multiprocessing and `Server.get_current()` is executed
    # in a spawned process because the server object is not initialized in that process.
    # It's unavoidable so ignored with this try-except, then developers have to
    # make sure not to use server-state in the spawned processes.
    logger.warning(
        "Failed to get the server object (%s). "
        "The server-state is not initialized in this process.",
        e,
    )

server_state: _ServerState
if _server:
    if hasattr(_server, _SERVER_STATE_KEY_):
        server_state = getattr(_server, _SERVER_STATE_KEY_)
    else:
        server_state = _ServerState()
        setattr(_server, _SERVER_STATE_KEY_, server_state)

server_state_lock: _ServerStateLock
if _server:
    if hasattr(_server, _SERVER_STATE_LOCK_KEY_):
        server_state_lock = getattr(_server, _SERVER_STATE_LOCK_KEY_)
    else:
        server_state_lock = _ServerStateLock(server_state=server_state)
        setattr(_server, _SERVER_STATE_LOCK_KEY_, server_state_lock)


force_rerun_bound_sessions = make_force_rerun_bound_sessions(server_state=server_state)


__all__ = [
    "server_state",
    "server_state_lock",
    "force_rerun_bound_sessions",
    "rerun_suppression",
    "no_rerun",
]



from .rerun_suppression import is_rerun_suppressed
from .server_state import ServerState


class RerunSuppressedError(Exception):
    pass


def make_force_rerun_bound_sessions(server_state: ServerState):
    def force_rerun_bound_sessions(key: str):
        if is_rerun_suppressed():
            raise RerunSuppressedError(
                "force_rerun_bound_sessions() cannot be used in the no_rerun context."
            )

        server_state.__items__[key]._rerun_bound_sessions()

    return force_rerun_bound_sessions



from typing import Any, Optional, Tuple, Union

ReprHash = Union[str, int]
ObjDictHash = Optional[str]
Hash = Tuple[ReprHash, ObjDictHash]


def calc_hash(val: Any) -> Hash:
    dict_hash: ObjDictHash = None
    if hasattr(val, "__dict__") and isinstance(val.__dict__, dict):
        dict_hash = repr(val.__dict__)

    repr_hash: ReprHash
    try:
        repr_hash = repr(val)
    except Exception:
        repr_hash = id(val)

    return (repr_hash, dict_hash)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['streamlit_server_state']

package_data = \
{'': ['*']}

install_requires = \
['packaging>=20.0', 'streamlit>=0.65.0']

setup_kwargs = {
    'name': 'streamlit-server-state',
    'version': '0.15.2',
    'description': '',
    'long_description': '# streamlit-server-state\nA "server-wide" state shared across the sessions.\n\n[![Tests](https://github.com/whitphx/streamlit-server-state/actions/workflows/tests.yml/badge.svg?branch=main)](https://github.com/whitphx/streamlit-server-state/actions/workflows/tests.yml?query=branch%3Amain)\n\n[![PyPI](https://img.shields.io/pypi/v/streamlit-server-state)](https://pypi.org/project/streamlit-server-state/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/streamlit-server-state)](https://pypi.org/project/streamlit-server-state/)\n[![PyPI - License](https://img.shields.io/pypi/l/streamlit-server-state)](https://pypi.org/project/streamlit-server-state/)\n[![PyPI - Downloads](https://img.shields.io/pypi/dm/streamlit-server-state)](https://pypi.org/project/streamlit-server-state/)\n\n[![ko-fi](https://ko-fi.com/img/githubbutton_sm.svg)](https://ko-fi.com/D1D2ERWFG)\n\n<a href="https://www.buymeacoffee.com/whitphx" target="_blank" rel="noreferrer"><img src="https://cdn.buymeacoffee.com/buttons/v2/default-yellow.png" alt="Buy Me A Coffee" width="180" height="50" ></a>\n\n[![GitHub Sponsors](https://img.shields.io/github/sponsors/whitphx?label=Sponsor%20me%20on%20GitHub%20Sponsors&style=social)](https://github.com/sponsors/whitphx)\n\n```python\nimport streamlit as st\n\nfrom streamlit_server_state import server_state, server_state_lock\n\nst.title("Global Counter Example")\n\nwith server_state_lock["count"]:  # Lock the "count" state for thread-safety\n    if "count" not in server_state:\n        server_state.count = 0\n\nincrement = st.button("Increment")\nif increment:\n    with server_state_lock.count:\n        server_state.count += 1\n\ndecrement = st.button("Decrement")\nif decrement:\n    with server_state_lock.count:\n        server_state.count -= 1\n\nst.write("Count = ", server_state.count)\n\n```\n\nAs above, the API is similar to [the built-in SessionState](https://blog.streamlit.io/session-state-for-streamlit/), except one major difference - a "lock" object.\nThe lock object is introduced for thread-safety because the server-state is accessed from multiple sessions, i.e. threads.\n\n## Auto-rerun\nWhen you assign a value to a server-state item, `server-state[key]`,\nserver-state automatically triggers re-running of all other sessions in which that server-state item is referred to so that all the references to the server-state return the latest value and all the sessions are kept up-to-date.\n\nFor example, with this mechanism, the [sample chat app (`app_chat.py`)](./app_chat.py) keeps showing the latest message list for all users.\n\n### Suppressing auto-rerun\n\nWhen this auto-rerun mechanism is not good for your use case, you can suppress auto-reruns upon the value assignments by using `no_rerun` context as below.\n```python\nfrom streamlit_server_state import server_state, no_rerun\n\n\nwith no_rerun:\n    server_state["foo"] = 42  # This does not trigger re-running of other sessions\n```\n\n### Manually trigger re-running\nUpon each value assignment, server-state checks whether the value has been changed and skips re-running if it has not for efficiency.\nThis works well in most cases, but it does not for example when the value is a complex mutable object and its field is mutated, while such usages are not recommended.\n\nAs exceptions, in such cases where the auto-rerun mechanism does not work well, you can manually trigger re-running by using `force_rerun_bound_sessions(key)`.\n\n```python\nif "foo" not in server_state:\n    server_state["foo"] = SomeComplexObject()\n\nserver_state["foo"].field = 42  # If this assignment does not trigger re-running,\n\nforce_rerun_bound_sessions("foo")  # You can do this.\n```\n\nBackground: https://discuss.streamlit.io/t/new-library-streamlit-server-state-a-new-way-to-share-states-across-sessions-on-the-server/14981/10\n\n## Examples\n* [`app_global_count`](./app_global_count.py): A sample app like [the official counter example for SessionState](https://blog.streamlit.io/session-state-for-streamlit/) which uses `streamlit-server-state` instead and the counter is shared among all the sessions on the server. This is a nice small example to see the usage and behavior of `streamlit-server-state`. Try to open the app in multiple browser tabs and see the counter is shared among them.\n* [`app_global_slider`](./app_global_slider.py): A slider widget (`st.slider`) whose value is shared among all sessions.\n* [`app_chat.py`](./app_chat.py): A simple chat app using `streamlit-server-state`.\n* [`app_chat_rooms.py`](./app_chat_rooms.py): A simple chat app with room separation.\n  [![Open in Streamlit](https://static.streamlit.io/badges/streamlit_badge_black_white.svg)](https://share.streamlit.io/whitphx/streamlit-server-state/main/app_chat_rooms.py)\n\n## Resources\n* [New library: streamlit-server-state, a new way to share states among the sessions on the server (Streamlit Community)](https://discuss.streamlit.io/t/new-library-streamlit-server-state-a-new-way-to-share-states-among-the-sessions-on-the-server/14981)\n',
    'author': 'Yuichiro Tachibana (Tsuchiya)',
    'author_email': 't.yic.yt@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/whitphx/streamlit-server-state',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7, !=2.7.*, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*, !=3.11.*',
}


setup(**setup_kwargs)
